#include <sstream>

#include <Windows.h>
#pragma comment(lib, "winmm.lib")

#include "pxcsensemanager.h"
#include "PXCFaceConfiguration.h"

#include <opencv2\opencv.hpp>

class RealSenseAsenseManager
{
public:

    ~RealSenseAsenseManager()
    {
        if ( senseManager != 0 ){
            senseManager->Release();
        }
    }

    void initilize()
    {
        // SenseManager𐶐
        senseManager = PXCSenseManager::CreateInstance();
        if ( senseManager == 0 ) {
            throw std::runtime_error( "SenseManager̐Ɏs܂" );
        }

        // J[Xg[Lɂ
        pxcStatus sts = senseManager->EnableStream( PXCCapture::StreamType::STREAM_TYPE_COLOR, COLOR_WIDTH, COLOR_HEIGHT, COLOR_FPS );
        if ( sts<PXC_STATUS_NO_ERROR ) {
            throw std::runtime_error( "J[Xg[̗LɎs܂" );
        }

		initializeFace();
    }

	void initializeFace(){

		// 猟oLɂ
		auto sts = senseManager->EnableFace();
		if (sts<PXC_STATUS_NO_ERROR) {
			throw std::runtime_error("猟o̗LɎs܂");
		}

		//猟o𐶐
		PXCFaceModule* faceModule = senseManager->QueryFace();
		if (faceModule == 0) {
			throw std::runtime_error("猟o̍쐬Ɏs܂");
		}

		//猟õvpeB擾
		PXCFaceConfiguration* config = faceModule->CreateActiveConfiguration();
		if (config == 0) {
			throw std::runtime_error("猟õvpeB擾Ɏs܂");
		}

		config->SetTrackingMode(PXCFaceConfiguration::TrackingModeType::FACE_MODE_COLOR_PLUS_DEPTH);
		config->ApplyChanges();


		// pCvC
		sts = senseManager->Init();
		if (sts<PXC_STATUS_NO_ERROR) {
			throw std::runtime_error("pCvC̏Ɏs܂");
		}

		// foCX̎擾
		auto device = senseManager->QueryCaptureManager()->QueryDevice();
		if (device == 0) {
			throw std::runtime_error("foCX̎擾Ɏs܂");
		}

		// ~[\ɂ
		device->SetMirrorMode(PXCCapture::Device::MirrorMode::MIRROR_MODE_HORIZONTAL);

		PXCCapture::DeviceInfo deviceInfo;
		device->QueryDeviceInfo(&deviceInfo);
		if (deviceInfo.model == PXCCapture::DEVICE_MODEL_IVCAM) {
			device->SetDepthConfidenceThreshold(1);
			device->SetIVCAMFilterOption(6);
			device->SetIVCAMMotionRangeTradeOff(21);
		}

		config->detection.isEnabled = true;
		config->landmarks.isEnabled = true;         //ǉFh}[N擾\ɂ
		config->landmarks.maxTrackedFaces = LANDMARK_MAXFACES;		//ǉFlɑΉ
		config->ApplyChanges();

		faceData = faceModule->CreateOutput();

	}

    void run()
    {
        // C[v
        while ( 1 ) {
            // t[f[^XV
            updateFrame();

            // \
            auto ret = showImage();
            if ( !ret ){
                break;
            }
        }
    }

private:

    void updateFrame()
    {
        // t[擾
        pxcStatus sts = senseManager->AcquireFrame( false );
        if ( sts < PXC_STATUS_NO_ERROR ) {
            return;
        }

		updateFaceFrame();

        // t[
        senseManager->ReleaseFrame();

        // t[[g\
        showFps();
    }

	void updateFaceFrame(){
		// t[f[^擾
		const PXCCapture::Sample *sample = senseManager->QuerySample();
		if (sample) {
			// ef[^\
			updateColorImage(sample->color);
		}

		//SenceManagerW[̊̃f[^XV
		faceData->Update();

		//o̐擾
		const int numFaces = faceData->QueryNumberOfDetectedFaces();

		//ǉF̃h}[Ni_j̃f[^̓ꕨp
		PXCFaceData::LandmarksData *landmarkData[LANDMARK_MAXFACES];
		PXCFaceData::LandmarkPoint* landmarkPoints;
		pxcI32 numPoints;

		//ꂼ̊炲Ƃɏ擾ѕ`揈s
		for (int i = 0; i < numFaces; ++i) {
			auto face = faceData->QueryFaceByIndex(i);
			if (face == 0){
				continue;
			}

			//̈ʒuf[^̓ꕨp
			PXCRectI32 faceRect = { 0 };


			//̈ʒu擾:ColorŎ擾
			auto detection = face->QueryDetection();
			if (detection != 0){
				detection->QueryBoundingRect(&faceRect);
			}

			//ǉFtFCXf[^烉h}[Ni_QjɂĂ̏𓾂
			landmarkData[i] = face->QueryLandmarks();
			if (landmarkData[i] != NULL)
			{
				//h}[Nf[^牽̓_Fł𒲂ׂ
				numPoints = landmarkData[i]->QueryNumPoints();
				//Fł_̐A_i[CX^X𐶐
				landmarkPoints = new PXCFaceData::LandmarkPoint[numPoints];
				//h}[Nf[^A_̈ʒu擾A\
				if (landmarkData[i]->QueryPoints(landmarkPoints)){
					for (int j = 0; j < numPoints; j++){
						{
							std::stringstream ss;
							ss << j ;
							//ss << landmarkPoints[j].source.alias;
							//int z = landmarkPoints[j].source.alias;
							cv::putText(colorImage, ss.str(), cv::Point(landmarkPoints[j].image.x, landmarkPoints[j].image.y), cv::FONT_HERSHEY_SIMPLEX, 0.5, cv::Scalar(0, 0, 255), 1, CV_AA);
						}
					}
				}

			}

		}

	}

    // J[摜XV
    void updateColorImage( PXCImage* colorFrame )
    {
        if ( colorFrame == 0 ){
            return;
        }
            
        PXCImage::ImageInfo info = colorFrame->QueryInfo();

        // f[^擾
        PXCImage::ImageData data;
        pxcStatus sts = colorFrame->AcquireAccess( PXCImage::Access::ACCESS_READ, PXCImage::PixelFormat::PIXEL_FORMAT_RGB24, &data );
        if ( sts < PXC_STATUS_NO_ERROR ) {
            throw std::runtime_error("J[摜̎擾Ɏs");
        }

        // f[^Rs[
        colorImage = cv::Mat( info.height, info.width, CV_8UC3 );
        memcpy( colorImage.data, data.planes[0], info.height * info.width * 3 );

        // f[^
        colorFrame->ReleaseAccess( &data );
    }

    // 摜\
    bool showImage()
    {
        // \
        cv::imshow( "Color Image", colorImage );

        int c = cv::waitKey( 10 );
        if ( (c == 27) || (c == 'q') || (c == 'Q') ){
            // ESC|q|Q for Exit
            return false;
        }

        return true;
    }

    // t[[g̕\
    void showFps()
    {
        static DWORD oldTime = ::timeGetTime();
        static int fps = 0;
        static int count = 0;

        count++;

        auto _new = ::timeGetTime();
        if ( (_new - oldTime) >= 1000 ){
            fps = count;
            count = 0;

            oldTime = _new;
        }

        std::stringstream ss;
        ss << fps;
        cv::putText( colorImage, ss.str(), cv::Point( 50, 50 ), cv::FONT_HERSHEY_SIMPLEX, 1.2, cv::Scalar( 0, 0, 255 ), 2, CV_AA );
    }

private:

    cv::Mat colorImage;
    PXCSenseManager* senseManager = 0;
    PXCFaceData* faceData = 0;

    const int COLOR_WIDTH = 640;
    const int COLOR_HEIGHT = 480;
    const int COLOR_FPS = 30;

	static const int LANDMARK_MAXFACES = 2;    //ǉF̃h}[N擾łőlݒ

};

void main()
{
    try{
        RealSenseAsenseManager asenseManager;
        asenseManager.initilize();
        asenseManager.run();
    }
    catch ( std::exception& ex ){
        std::cout << ex.what() << std::endl;
    }
}
